@@ -110,28 +110,36 @@
 use std::ffi::{CStr, CString};
 use std::io::Read;
 use std::io::Error as IoError;
-use std::borrow::Borrow;
+use std::borrow::{Borrow, Cow};
 use std::num::NonZeroI32;
 use std::error::Error;
 use std::fmt;
-use std::convert::{From, TryInto};
-use std::io;
+use std::io::{self, Write};
 
 pub use any::{AnyHashableLuaValue, AnyLuaString, AnyLuaValue};
-pub use functions_write::{Function, InsideCallback};
-pub use functions_write::{function0, function1, function2, function3, function4, function5};
-pub use functions_write::{function6, function7, function8, function9, function10};
+pub use functions_write::{
+    Function, InsideCallback,
+    function0, function1, function2, function3, function4, function5,
+    function6, function7, function8, function9, function10,
+    protected_call,
+};
 pub use lua_functions::LuaFunction;
-pub use lua_functions::LuaFunctionCallError;
 pub use lua_functions::{LuaCode, LuaCodeFromReader};
-pub use lua_tables::{LuaTable, LuaTableIterator, MethodCallError};
-pub use rust_tables::PushIterError;
+pub use lua_tables::{LuaTable, LuaTableIterator};
+pub use object::{
+    Call, Callable, CallError,
+    Index, Indexable, IndexableRW, MethodCallError, NewIndex,
+    Object,
+};
+pub use rust_tables::{TableFromIter, PushIterError, PushIterErrorOf};
 pub use tuples::TuplePushError;
 pub use userdata::UserdataOnStack;
 pub use userdata::{push_userdata, read_userdata, push_some_userdata};
 pub use values::{StringInLua, Nil, Null, True, False, Typename, ToString};
 pub use ::tlua_derive::*;
 
+#[deprecated = "Use `CallError` instead"]
+pub type LuaFunctionCallError<E> = CallError<E>;
 pub type LuaTableMap = std::collections::HashMap<AnyHashableLuaValue, AnyLuaValue>;
 pub type LuaSequence = Vec<AnyLuaValue>;
 
@@ -142,16 +150,37 @@ mod functions_write;
 mod lua_functions;
 mod lua_tables;
 mod macros;
+mod object;
 mod rust_tables;
 mod userdata;
+pub mod util;
 mod values;
 mod tuples;
 
 pub type LuaState = *mut ffi::lua_State;
 
+/// A static lua context that must be created from an existing lua state pointer
+/// and that will **not** be closed when dropped.
+pub type StaticLua = Lua<on_drop::Ignore>;
+
+/// A temporary lua context that will be closed when dropped.
+pub type TempLua = Lua<on_drop::Close>;
+
+/// A lua context corresponding to a lua thread (see [`ffi::lua_newthread`])
+/// stored in the global [REGISTRY](ffi::LUA_REGISTRYINDEX) and will be removed
+/// from there when dropped.
+///
+/// `LuaThread` currently can only be created from an instance of [`StaticLua`]
+/// because closing a state from which a thread has been created is forbidden.
+pub type LuaThread = Lua<on_drop::Unref>;
+
 /// Main object of the library.
 ///
-/// The lifetime parameter corresponds to the lifetime of the content of the Lua context.
+/// The type parameter `OnDrop` specifies what happens with the underlying lua
+/// state when the instance gets dropped. There are currently 3 supported cases:
+/// - `on_drop::Ignore`: nothing happens
+/// - `on_drop::Close`: [`ffi::lua_close`] is called
+/// - `on_drop::Unref`: [`ffi::luaL_unref`] is called with the associated value
 ///
 /// # About panic safety
 ///
@@ -159,9 +188,48 @@ pub type LuaState = *mut ffi::lua_State;
 /// then it will probably stay in a corrupt state. Trying to use the `Lua` again will most likely
 /// result in another panic but shouldn't result in unsafety.
 #[derive(Debug)]
-pub struct Lua {
-    lua: *mut ffi::lua_State,
-    must_be_closed: bool,
+pub struct Lua<OnDrop>
+where
+    OnDrop: on_drop::OnDrop,
+{
+    lua: LuaState,
+    on_drop: OnDrop,
+}
+
+mod on_drop {
+    use crate::{LuaState, ffi};
+
+    pub trait OnDrop {
+        fn on_drop(&mut self, l: LuaState);
+    }
+
+    /// See [`StaticLua`].
+    #[derive(Debug)]
+    pub struct Ignore;
+
+    impl OnDrop for Ignore {
+        fn on_drop(&mut self, _: LuaState) {}
+    }
+
+    /// See [`TempLua`].
+    #[derive(Debug)]
+    pub struct Close;
+
+    impl OnDrop for Close {
+        fn on_drop(&mut self, l: LuaState) {
+            unsafe { ffi::lua_close(l) }
+        }
+    }
+
+    /// See [`LuaThread`].
+    #[derive(Debug)]
+    pub struct Unref(pub i32);
+
+    impl OnDrop for Unref {
+        fn on_drop(&mut self, l: LuaState) {
+            unsafe { ffi::luaL_unref(l, ffi::LUA_REGISTRYINDEX, self.0) }
+        }
+    }
 }
 
 /// RAII guard for a value pushed on the stack.
@@ -200,7 +268,9 @@ impl<L: AsLua> PushGuard<L> {
     /// Creates a new `PushGuard` from this Lua context representing `size` items on the stack.
     /// When this `PushGuard` is destroyed, `size` items will be popped.
     ///
-    /// This is unsafe because the Lua stack can be corrupted if this is misused.
+    /// # Safety
+    /// There must be at least `size` elements on the `lua` stack and it must be
+    /// safe to drop them at the same time with this `PushGuard`.
     #[inline]
     pub unsafe fn new(lua: L, size: i32) -> Self {
         PushGuard {
@@ -225,7 +295,9 @@ impl<L: AsLua> PushGuard<L> {
     /// Prevents the value from being popped when the `PushGuard` is destroyed, and returns the
     /// number of elements on the Lua stack.
     ///
-    /// This is unsafe because the Lua stack can be corrupted if this is misused.
+    /// # Safety
+    /// The values on the stack will not be popped automatically, so the caller
+    /// must ensure nothing is leaked.
     #[inline]
     pub unsafe fn forget(self) -> i32 {
         self.forget_internal()
@@ -378,16 +450,13 @@ pub trait AsLua {
     ///
     /// If `I::Item` pushes more than 2 values or an error happens during an
     /// attempt to push, the function returns `Err((e, self))` where `e` is a
-    /// `PushIterError`.
+    /// `PushIterErrorOf`.
     ///
     /// Returns a `PushGuard` which captures `self` by value and stores the
     /// amount of values pushed onto the stack (exactly 1 -- lua table).
     #[inline(always)]
     fn try_push_iter<I>(self, iterator: I)
-        -> Result<
-            PushGuard<Self>,
-            (PushIterError<<<I as Iterator>::Item as PushInto<LuaState>>::Err>, Self)
-        >
+        -> Result<PushGuard<Self>, (PushIterErrorOf<I>, Self)>
     where
         Self: Sized,
         I: Iterator,
@@ -422,18 +491,34 @@ pub trait AsLua {
     {
         T::lua_read_at_position(self, index)
     }
+
+    /// Call a rust function in protected mode. If a lua error is thrown during
+    /// execution of `f` the function will return a `LuaError`.
+    ///
+    /// This can also be sometimes used to catch other C++ exceptions although
+    /// be careful with that.
+    #[inline(always)]
+    fn pcall<F, R>(&self, f: F) -> Result<R, LuaError>
+    where
+        F: FnOnce(StaticLua) -> R,
+    {
+        protected_call(self, f)
+    }
 }
 
 impl<T> AsLua for &'_ T
 where
-    T: AsLua,
+    T: ?Sized + AsLua,
 {
     fn as_lua(&self) -> *mut ffi::lua_State {
         T::as_lua(self)
     }
 }
 
-impl AsLua for Lua {
+impl<D> AsLua for Lua<D>
+where
+    D: on_drop::OnDrop,
+{
     #[inline]
     fn as_lua(&self) -> *mut ffi::lua_State {
         self.lua
@@ -635,23 +720,33 @@ pub enum LuaError {
 
     /// There was an error during execution of the Lua code
     /// (for example not enough parameters for a function call).
-    ExecutionError(String),
+    ExecutionError(Cow<'static, str>),
 
     /// There was an IoError while reading the source code to execute.
     ReadError(IoError),
 
     /// The call to `eval` has requested the wrong type of data.
     WrongType{
+        when: &'static str,
         rust_expected: String,
         lua_actual: String,
     },
 }
 
 impl LuaError {
-    pub fn wrong_type<T, L: AsLua>(lua: L, n_values: i32) -> Self {
+    pub fn wrong_type_returned<T, L: AsLua>(lua: L, n_values: i32) -> Self {
+        Self::wrong_type::<T, _>(lua, n_values, "Wrong type returned by Lua")
+    }
+
+    pub fn wrong_type_passed<T, L: AsLua>(lua: L, n_values: i32) -> Self {
+        Self::wrong_type::<T, _>(lua, n_values, "Wrong type passed into rust callback")
+    }
+
+    pub fn wrong_type<T, L: AsLua>(lua: L, n_values: i32, when: &'static str) -> Self {
         let nz = unsafe { NonZeroI32::new_unchecked(-n_values) };
         let start = AbsoluteIndex::new(nz, lua.as_lua());
         Self::WrongType {
+            when,
             rust_expected: std::any::type_name::<T>().into(),
             lua_actual: typenames(lua, start, n_values as _),
         }
@@ -677,29 +772,28 @@ pub fn typenames(lua: impl AsLua, start: AbsoluteIndex, count: u32) -> String {
         _ => {}
     }
 
-    let mut res = vec![std::borrow::Cow::Borrowed("(")];
+    let mut res = Vec::with_capacity(32);
+    write!(res, "(").expect("writing to vec cannot fail");
     let end = start + count - 1;
     for i in start..end {
-        res.push(single_typename(i));
-        res.push(", ".into());
+        write!(res, "{}, ", single_typename(i)).expect("writing to vec cannot fail");
     }
-    res.push(single_typename(end));
-    res.push(")".into());
-    res.join("")
+    write!(res, "{})", single_typename(end)).expect("writing to vec cannot fail");
+    // concatenation of utf8 is utf8
+    unsafe { String::from_utf8_unchecked(res) }
 }
 
 impl fmt::Display for LuaError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         use LuaError::*;
 
-        match *self {
-            SyntaxError(ref s) => write!(f, "Syntax error: {}", s),
-            ExecutionError(ref s) => write!(f, "Execution error: {}", s),
-            ReadError(ref e) => write!(f, "Read error: {}", e),
-            WrongType{
-                rust_expected: ref e1,
-                lua_actual: ref e2
-            } => write!(f, "Wrong type returned by Lua: {} expected, got {}", e1, e2),
+        match self {
+            SyntaxError(s) => write!(f, "Syntax error: {}", s),
+            ExecutionError(s) => write!(f, "Execution error: {}", s),
+            ReadError(e) => write!(f, "Read error: {}", e),
+            WrongType { when, rust_expected, lua_actual } => {
+                write!(f, "{}: {} expected, got {}", when, rust_expected, lua_actual)
+            }
         }
     }
 }
@@ -709,10 +803,10 @@ impl Error for LuaError {
         use LuaError::*;
 
         match *self {
-            SyntaxError(ref s) => &s,
-            ExecutionError(ref s) => &s,
+            SyntaxError(ref s) => s,
+            ExecutionError(ref s) => s,
             ReadError(_) => "read error",
-            WrongType{rust_expected: _, lua_actual: _} => "wrong type returned by Lua",
+            WrongType { when, .. } => when,
         }
     }
 
@@ -723,7 +817,7 @@ impl Error for LuaError {
             SyntaxError(_) => None,
             ExecutionError(_) => None,
             ReadError(ref e) => Some(e),
-            WrongType{rust_expected: _, lua_actual: _} => None,
+            WrongType { .. } => None,
         }
     }
 }
@@ -734,8 +828,8 @@ impl From<io::Error> for LuaError {
     }
 }
 
-impl Lua {
-    /// Builds a new empty Lua context.
+impl TempLua {
+    /// Builds a new empty TempLua context.
     ///
     /// There are no global variables and the registry is totally empty. Even the functions from
     /// the standard library can't be used.
@@ -755,7 +849,7 @@ impl Lua {
     /// The function panics if the underlying call to `lua_newstate` fails
     /// (which indicates lack of memory).
     #[inline]
-    pub fn new() -> Lua {
+    pub fn new() -> Self {
         let lua = unsafe { ffi::luaL_newstate() };
         if lua.is_null() {
             panic!("lua_newstate failed");
@@ -771,24 +865,59 @@ impl Lua {
 
         unsafe { ffi::lua_atpanic(lua, panic) };
 
-        Lua {
-            lua,
-            must_be_closed: true,
-        }
+        unsafe { Self::from_existing(lua) }
     }
 
-    /// Takes an existing `lua_State` and build a Lua object from it.
+    /// Takes an existing `lua_State` and build a TemplLua object from it.
     ///
-    /// If `close_at_the_end` is true, `lua_close` will be called on the `lua_State` in the
-    /// destructor.
+    /// `lua_close` will be called on the `lua_State` in drop.
+    ///
+    /// # Safety
+    /// A pointer to a valid `lua` context must be provided which is ok to be
+    /// closed.
     #[inline]
-    pub unsafe fn from_existing_state<T>(lua: *mut T, close_at_the_end: bool) -> Lua {
-        Lua {
-            lua: std::mem::transmute(lua),
-            must_be_closed: close_at_the_end,
+    pub unsafe fn from_existing<T>(lua: *mut T) -> Self {
+        Self {
+            lua: lua as _,
+            on_drop: on_drop::Close,
+        }
+    }
+}
+
+impl StaticLua {
+    /// Takes an existing `lua_State` and build a StaticLua object from it.
+    ///
+    /// `lua_close` is **NOT** called in `drop`.
+    ///
+    /// # Safety
+    /// A pointer to a valid `lua` context must be provided.
+    #[inline]
+    pub unsafe fn from_static<T>(lua: *mut T) -> Self {
+        Self {
+            lua: lua as _,
+            on_drop: on_drop::Ignore,
         }
     }
 
+    /// Creates a new Lua thread with an independent stack and runs the provided
+    /// function within it. The new state has access to all the global objects
+    /// available to `self`.
+    pub fn new_thread(&self) -> LuaThread {
+        unsafe {
+            let lua = ffi::lua_newthread(self.as_lua());
+            let r = ffi::luaL_ref(self.as_lua(), ffi::LUA_REGISTRYINDEX);
+            LuaThread {
+                lua,
+                on_drop: on_drop::Unref(r)
+            }
+        }
+    }
+}
+
+impl<OnDrop> Lua<OnDrop>
+where
+    OnDrop: on_drop::OnDrop,
+{
     /// Opens all standard Lua libraries.
     ///
     /// See the reference for the standard library here:
@@ -942,7 +1071,7 @@ impl Lua {
     /// ```
     #[inline(always)]
     // TODO(gmoshkin): this method should be part of AsLua
-    pub fn exec<'lua>(&'lua self, code: &str) -> Result<(), LuaError> {
+    pub fn exec(&self, code: &str) -> Result<(), LuaError> {
         LuaFunction::load(self, code)?
             .into_call()
     }
@@ -997,7 +1126,7 @@ impl Lua {
     /// ```
     #[inline(always)]
     // TODO(gmoshkin): this method should be part of AsLua
-    pub fn exec_from<'lua>(&'lua self, code: impl Read) -> Result<(), LuaError> {
+    pub fn exec_from(&self, code: impl Read) -> Result<(), LuaError> {
         LuaFunction::load_from_reader(self, code)?
             .into_call()
     }
@@ -1142,7 +1271,7 @@ impl Lua {
     /// ```
     #[inline]
     // TODO(gmoshkin): this method should be part of AsLua
-    pub fn empty_array<'lua, I>(&'lua self, index: I) -> LuaTable<PushGuard<&'lua Self>>
+    pub fn empty_array<I>(&self, index: I) -> LuaTable<PushGuard<&Self>>
     where
         I: Borrow<str>,
     {
@@ -1200,7 +1329,7 @@ impl Lua {
     /// ```
     #[inline]
     // TODO(gmoshkin): this method should be part of AsLua
-    pub fn globals_table<'lua>(&'lua self) -> LuaTable<PushGuard<&'lua Self>> {
+    pub fn globals_table(&self) -> LuaTable<PushGuard<&Self>> {
         unsafe {
             ffi::lua_pushglobaltable(self.lua);
             let guard = PushGuard::new(self, 1);
@@ -1209,12 +1338,19 @@ impl Lua {
     }
 }
 
-impl Drop for Lua {
+impl Default for TempLua {
+    fn default() -> Self {
+        Self::new()
+    }
+}
+
+impl<T> Drop for Lua<T>
+where
+    T: on_drop::OnDrop,
+{
     #[inline]
     fn drop(&mut self) {
-        if self.must_be_closed {
-            unsafe { ffi::lua_close(self.lua) }
-        }
+        self.on_drop.on_drop(self.lua)
     }
 }
 
@@ -1239,14 +1375,16 @@ impl AbsoluteIndex {
     {
         let top = unsafe { ffi::lua_gettop(lua.as_lua()) };
         let i = index.get();
-        if unsafe { ffi::is_relative_index(i) } {
-            let index = (top + i + 1).try_into().expect("Invalid relative index");
-            Self(NonZeroI32::new(index).expect("Invalid relative index"))
+        if ffi::is_relative_index(i) {
+            Self(NonZeroI32::new(top + i + 1).expect("Invalid relative index"))
         } else {
             Self(index)
         }
     }
 
+    /// # Safety
+    /// `index` must be a valid absolute or relative index into the lua stack
+    /// with which it's going to be used
     pub unsafe fn new_unchecked(index: NonZeroI32) -> Self {
         Self(index)
     }
