@@ -1,5 +1,6 @@
 use std::convert::TryFrom;
 use std::borrow::Cow;
+use std::ffi::{CStr, CString};
 use std::mem::MaybeUninit;
 use std::num::NonZeroI32;
 use std::slice;
@@ -250,6 +251,20 @@ impl_push_read!{ String,
     }
 }
 
+impl_push_read!{ CString,
+    push_to_lua(&self, lua) {
+        push_string_impl!(self, lua)
+    }
+    push_into_lua(self, lua) {
+        push_string_impl!(self, lua)
+    }
+    read_at_position(lua, index) {
+        lua_read_string_impl!(lua, index,
+            |slice: &[u8], lua| CString::new(slice).map_err(|_| lua)
+        )
+    }
+}
+
 impl_push_read!{ AnyLuaString,
     push_to_lua(&self, lua) {
         push_string_impl!(self, lua)
@@ -270,6 +285,19 @@ impl_push_read!{ str,
     }
 }
 
+impl_push_read!{ CStr,
+    push_to_lua(&self, lua) {
+        unsafe {
+            ffi::lua_pushlstring(
+                lua.as_lua(),
+                self.as_ptr() as _,
+                self.to_bytes().len() as _,
+            );
+            Ok(PushGuard::new(lua, 1))
+        }
+    }
+}
+
 /// String on the Lua stack.
 ///
 /// It is faster -but less convenient- to read a `StringInLua` rather than a `String` because you
@@ -325,7 +353,7 @@ where
             |slice: &'a [u8], lua|
                 match str::from_utf8(slice) {
                     Ok(str_ref) => Ok(StringInLua { lua, str_ref }),
-                    Err(_) => return Err(lua)
+                    Err(_) => Err(lua)
                 }
         )
     }
@@ -336,7 +364,7 @@ impl<'a, L> Deref for StringInLua<'a, L> {
 
     #[inline]
     fn deref(&self) -> &str {
-        &self.str_ref
+        self.str_ref
     }
 }
 
@@ -351,7 +379,7 @@ impl_push_read!{ bool,
         }
     }
     read_at_position(lua, index) {
-        if unsafe { ffi::lua_isboolean(lua.as_lua(), index.into()) } != true {
+        if !unsafe { ffi::lua_isboolean(lua.as_lua(), index.into()) } {
             return Err(lua);
         }
 
@@ -535,6 +563,8 @@ impl_push_read!{Null,
 }
 
 #[derive(Copy, Clone, Debug, PartialEq, Eq, Hash, PartialOrd, Ord)]
+#[derive(serde::Serialize, serde::Deserialize)]
+#[serde(try_from = "bool", into = "bool")]
 pub struct True;
 
 impl From<True> for bool {
@@ -569,7 +599,16 @@ impl_push_read!{True,
     }
 }
 
+impl std::fmt::Display for True {
+    #[inline(always)]
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        bool::from(*self).fmt(f)
+    }
+}
+
 #[derive(Copy, Clone, Debug, PartialEq, Eq, Hash, PartialOrd, Ord)]
+#[derive(serde::Serialize, serde::Deserialize)]
+#[serde(try_from = "bool", into = "bool")]
 pub struct False;
 
 impl From<False> for bool {
@@ -604,6 +643,13 @@ impl_push_read!{False,
     }
 }
 
+impl std::fmt::Display for False {
+    #[inline(always)]
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        bool::from(*self).fmt(f)
+    }
+}
+
 #[derive(Copy, Clone, Debug, PartialEq, Eq, Hash, PartialOrd, Ord)]
 pub struct Typename(pub &'static str);
 
@@ -635,6 +681,12 @@ impl From<ToString> for String {
     }
 }
 
+impl<'a> From<ToString> for Cow<'a, str> {
+    fn from(other: ToString) -> Self {
+        Cow::Owned(other.0)
+    }
+}
+
 impl std::fmt::Display for ToString {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
         write!(f, "{}", self.0)
